@@ -67,6 +67,7 @@ WORD machine = IMAGE_FILE_MACHINE_AMD64;
 WORD machine = IMAGE_FILE_MACHINE_I386;
 #endif
 ULONG64 image_base = 0;
+ULONG64 low_addr;
 BOOL down_flag = FALSE;
 BOOL image_info_flag = FALSE;
 BOOL image_storage_flag = FALSE;
@@ -116,6 +117,30 @@ ULONG cygwin_dll_image_size = 0;
 
 #define LONG_PATH_MAX 32768
 
+int
+check_base_address_sanity (ULONG64 addr, BOOL at_start)
+{
+#if defined(__CYGWIN__) || defined(__MSYS__)
+  /* Sanity checks for Cygwin:
+   *
+   * - No DLLs below 0x38000000 on 32 bit, W10 1703+ rebase those on
+   *   runtime anyway
+   * - No DLLs below 0x2:00000000, ever, on 64 bit.
+   */
+  if (addr <= low_addr)
+    {
+      if (at_start)
+	fprintf (stderr, "%s: Invalid Baseaddress 0x%" PRIx64 ", must be > 0x%" PRIx64 "\n",
+		 progname, (uint64_t) addr, (uint64_t) low_addr);
+      else
+	fprintf (stderr, "%s: Too many DLLs for available address space: %s\n",
+		 progname, strerror (ENOMEM));
+      return -1;
+    }
+#endif
+  return 0;
+}
+
 void
 gen_progname (const char *arg0)
 {
@@ -243,7 +268,8 @@ main (int argc, char *argv[])
       /* Rebase with database support. */
       BOOL header;
 
-      merge_image_info ();
+      if (merge_image_info () < 0)
+	return 2;
       status = TRUE;
       for (i = 0; i < img_info_size; ++i)
 	if (img_info_list[i].flag.needs_rebasing)
@@ -742,29 +768,36 @@ merge_image_info ()
   end = img_info_size - 1;
   while (img_info_list[0].base == 0)
     {
-      ULONG64 new_base;
+      ULONG64 new_base = 0;
 
       /* Skip trailing entries as long as there is no hole. */
-       while (img_info_list[end].base + img_info_list[end].slot_size + offset
-	     >= floating_image_base)
+       while (end > 0
+	      && img_info_list[end].base + img_info_list[end].slot_size
+		 + offset >= floating_image_base)
 	{
 	  floating_image_base = img_info_list[end].base;
 	  --end;
 	}
+
       /* Test if one of the DLLs with address 0 fits into the hole. */
-      for (i = 0, new_base = 0; img_info_list[i].base == 0; ++i, new_base = 0)
+      for (i = 0; img_info_list[i].base == 0; ++i)
 	{
-	  new_base = floating_image_base - img_info_list[i].slot_size - offset;
-	  if (new_base >= img_info_list[end].base
-			  + img_info_list[end].slot_size
+	  ULONG64 base = floating_image_base - img_info_list[i].slot_size
+		  - offset;
+	  /* Check if address is still valid */
+	  if (check_base_address_sanity (base, FALSE))
+	    return -1;
+	  if (base >= img_info_list[end].base + img_info_list[end].slot_size
 #if defined(__CYGWIN__) || defined(__MSYS__)
 	      /* Don't overlap the Cygwin/MSYS DLL. */
-	      && (new_base >= cygwin_dll_image_base + cygwin_dll_image_size
-		  || new_base + img_info_list[i].slot_size
-		     <= cygwin_dll_image_base)
+	      && (base >= cygwin_dll_image_base + cygwin_dll_image_size
+		  || base + img_info_list[i].slot_size <= cygwin_dll_image_base)
 #endif
 	     )
-	    break;
+	    {
+	      new_base = base;
+	      break;
+	    }
 	}
       /* Found a match.  Mount into list. */
       if (new_base)
@@ -781,12 +814,18 @@ merge_image_info ()
 #if defined(__CYGWIN__) || defined(__MSYS__)
       if (floating_image_base >= cygwin_dll_image_base + cygwin_dll_image_size
 	  && img_info_list[end].base < cygwin_dll_image_base)
-	floating_image_base = cygwin_dll_image_base;
+	  floating_image_base = cygwin_dll_image_base;
       else
 #endif
 	{
 	  floating_image_base = img_info_list[end].base;
-	  --end;
+	  if (--end < 0)
+	    {
+	      fprintf (stderr,
+		       "%s: Too many DLLs for available address space: %s\n",
+		       progname, strerror (ENOMEM));
+	      return -1;
+	    }
 	}
     }
 
@@ -1209,6 +1248,16 @@ parse_args (int argc, char *argv[])
       exit (1);
     }
 
+  /* The low address for 32 bit is extremly low, and apparently
+     W10 1703 and later rebase all DLLs with start addresses < 0x38000000
+     at runtime.  However, we have so many DLLs that a hardcoded lowest
+     address of 0x38000000 is just not feasible. */
+  low_addr = (machine == IMAGE_FILE_MACHINE_I386) ? 0x001000000ULL
+						  : 0x200000000ULL;
+
+  if (image_base && check_base_address_sanity (image_base, TRUE) < 0)
+    exit (1);
+
   args_index = optind;
 
   /* Initialize db_file and tmp_file from pattern */
